
local configuration = {}

local safer = require("safer")
local schema = require("schema")
local util = require("sga.util")


local function get(config_data, filename, config_default)
   local config = {}
   config_default.sgas = nil -- avoids multiple inside multiple
   config_default.__index = config_default -- sets default values for the configuration
   config_default.__pairs = util.metatable_pairs -- forces the iterator to go through default values
   setmetatable(config, config_default)
   local chunk, err = load(config_data, filename, "t", config)
   if not chunk then
      return nil, "Failed loading config file "..filename..": "..err
   end
   config.os = { getenv = os.getenv }
   config.tonumber = tonumber

   local ok, err = pcall(chunk)
   config.os = nil
   config.tonumber = nil
   if not ok then
      return nil, "Failed processing config file "..filename..": "..err
   end

   local warning = nil
   -- Warns if the sga_name is being inherited from the default configuration and not explicitly declared
   if config_default["sga_name"] and not rawget(config, "sga_name") then
      warning = string.format("[WARNING] Config file %s contains configuration using default sga_name %s", filename, config_default["sga_name"])
   end
   return config, warning
end

local function get_from_multiple(filename, sga_name, config)
   local err = "Failed getting config from multiple config file %s: %s"
   if not sga_name then
      return nil, string.format(err, filename, "missing sga_name argument")
   end
   local sgas_data = config.sgas
   if not sgas_data then
      return nil, string.format(err, filename, "not a multiple config file")
   end

   -- Indexes all sga configs by sga_name
   local sgas = {}
   local warning = {}
   for key, data in ipairs(sgas_data) do
      local sga_config, err_msg = {}
      sga_config, err_msg = get(data, filename, config)
      if err_msg then
         table.insert(warning, err_msg)
      end
      if sga_config then
         configuration.check(sga_config, true)
         if sga_config.sga_name then
            if sgas[sga_config.sga_name] then
               if sga_name == sga_config.sga_name then
                  return nil, string.format(err, filename, "multiple definition of sga '"..sga_name.."'")
               end
               table.insert(warning,
                  string.format("[WARNING] Config file %s has duplicate configuration for sga '%s'", filename, sga_config.sga_name))
            end
            sgas[sga_config.sga_name] = sga_config
         end
      else
         table.insert(warning,
            string.format("[WARNING] Config file %s has malformed configuration:\n %s", filename, data))
      end
   end
   warning = table.concat(warning, '\n')

   local config_data = sgas[sga_name]
   if not config_data then
      return nil, warning..'\n'..string.format(err, filename, "'"..sga_name.."' not found")
   end
   return config_data, warning
end

function configuration.read(filename, sga_name)
   local config_fd = io.open(filename, "r")
   if not config_fd then
      return nil, "Could not read config file "..filename
   end
   local config_data = config_fd:read("*a")
   config_fd:close()

   local config, err = {}
   config, err = get(config_data, filename, config)
   if not config then
      return nil, err
   end

   local is_multiple = (config.sgas and #config.sgas > 0)
   if sga_name and not is_multiple then
      return nil, string.format(
         "Failed getting config '%s': file %s is not a valid multiple config file",
         sga_name, filename)
   end
   local message = nil
   if is_multiple then
      config, message = get_from_multiple(filename, sga_name, config)
      if not config then
         return nil, message
      end
   end

   config = safer.readonly(config)
   return config, message
end

function configuration.check(config, just_warn)
   local config_schema = schema.Record {
      csbase_server = schema.Pattern("https?://.*"),
      platform = schema.String,
      sgad_bind_addr = schema.Optional(schema.String),
      sgad_host = schema.String,
      sgad_port = schema.NumberFrom(1,65535),
      sga_name = schema.Pattern("[A-Za-z0-9_]+"),
      sga_type = schema.Optional(schema.String),
      status_interval_s = schema.PositiveNumber,
      exec_polling_interval_s = schema.PositiveNumber,
      register_retry_s = schema.PositiveNumber,
      driver = schema.Test(function(v)
         local ok, err = pcall(require, v)
         if not ok then
            io.stderr:write("Failed loading driver '"..tostring(v).."': "..err.."\n")
         end
         return ok
      end, "must be a valid module"),
      driver_config = schema.Optional(schema.Map(schema.String, schema.Any)),
      resources = schema.Optional(schema.Collection(schema.String)),
      extra_config = schema.Optional(schema.Map(schema.String, schema.Any)),
      -- TODO validate path in *_dir parameters:
      runtime_data_dir = schema.String,
      project_root_dir = schema.String,
      algorithm_root_dir = schema.String,
      sandbox_root_dir = schema.String,
   }
   local err = schema.CheckSchema(config, config_schema)
   if err then
      err = " Configuration error: \n"..tostring(err).."\n"
      if config.sga_name then
         err =  "["..config.sga_name.."]"..err
      end
      if just_warn then
         io.stderr:write("[WARNING]"..err)
      else
         io.stderr:write(err)
         os.exit(1)
      end
   end
end

return configuration
