
local joblist = {}

local safer = require("safer")
local copas = require("copas")
local serpent = require("serpent")
local util = require("sga.util")

local jobs = {}
local attempted_jobs = {}

function joblist:load()
   local data, err = util.read_file(self.filename)
   if not data then
      return nil, err
   end
   local ok, newjobs = serpent.load(data)
   if not data then
      return nil, err
   end
   return newjobs
end

function joblist:save()
   util.write_file(self.filename, serpent.dump(jobs))
end

function joblist:new_job(cmd_id, cmd_string, parameters)
   local jid = self.next_jid
   while jobs[jid] do
      jid = jid + 1
   end
   self.next_jid = jid + 1

   local job = {
      jid = jid,
      cmd_id = cmd_id,
      cmd_string = cmd_string,
      -- safer.readonly doesn't work with the parameters table
      -- parameters = safer.readonly(parameters),
      parameters = parameters,
      data = {},
      sandboxes = util.convert_to_list(parameters, "csbase_command_sandbox_paths")
   }
   attempted_jobs[jid] = job
   return safer.readonly(job)
end

function joblist.accept_job(_, job, accepted)
   if accepted then
      jobs[job.jid] = attempted_jobs[job.jid]
   else
      attempted_jobs[job.jid] = nil
   end
end

function joblist.each()
   local key = next(jobs)
   return function()
      if not key then
         return nil
      end
      local val = jobs[key]
      key = next(jobs, key)
      return safer.readonly(val)
   end
end

function joblist.get(_, jid)
   return safer.readonly(jobs[jid])
end

function joblist:finish(job)
   job = jobs[job.jid]
   if job then
      job.done = os.time()
      self:save()
      return true
   end
   return false
end

function joblist:terminate()
   self:save()
   self.terminated = true
   if self.gc_coro then
      copas.wakeup(self.gc_coro)
   end
end

function joblist.new(config)
   local self = {
      filename = config.runtime_data_dir.."/sga_jobs_persisted_data."..config.driver,
      terminated = false,
      gc_coro = false,
      next_jid = 0,
   }
   setmetatable(self, { __index = joblist })

   local newjobs, err = self:load()
   if newjobs then
      for jid, _ in pairs(newjobs) do
         if jid > self.next_jid then
            self.next_jid = jid + 1
         end
      end
      jobs = newjobs
   end

   -- Jobs garbage collector:
   -- keep jobs in joblist for a minute after they are done,
   -- so we can still respond to queries about them during this period.
   self.gc_coro = copas.addthread(function()
      while not self.terminated do
         copas.sleep(5)
         if self.terminated then break end
         local now = os.time()
         for jid, job in pairs(jobs) do
            if job.done and now > job.done + 60 then
               jobs[jid] = nil
               self:save()
            end
         end
      end
   end)

   return safer.table(self)
end

return joblist
