
local util = {}

local lfs = require("lfs")
local stat = require("posix.sys.stat")

local format_str = function(s)
   s = tostring(s)
   local _, n = s:gsub('\n','')
   if n > 0 then
      return "\n[[\n"..s.."]]" -- block string
   else
      return s -- line string
   end
end

function util.write_file(name, data)
   local fd = io.open(name, "w")
   if not fd then
      return nil, "could not write "..name
   end
   fd:write(data)
   fd:close()
   return true
end

function util.read_file(name, offset)
   offset = offset or 0
   local fd, err = io.open(name, "r")
   if not fd then
      return nil, "could not read "..name.." : "..err
   end
   local n_bytes, err = fd:seek("set", offset)
   if not n_bytes then
      return nil, "could not read "..name.." : "..err
   end
   local data = fd:read("*a")
   fd:close()
   return data
end

function util.metatable_pairs(tbl)

   -- Iterator function that goes through both the table and its metatable keys
   local function stateless_iter(tbl, k)
      local v
      if k == nil or rawget(tbl,k) then
         k, v = next(tbl, k)
      end
      if nil~=v then
         return k,v
      else
         local mt = getmetatable(tbl)
         if mt and mt.__index and type(mt.__index) == "table" then
            while true do
               k, v = next(mt, k)
               if not v then
                  break
               end
               local is_metamethod = (k and k:sub(1, 2) == "__")
               local is_from_original = rawget(tbl, k)
               if not is_metamethod and not is_from_original then
                  return k, v
               end
            end
         end
       end
    end
  
    -- Return an iterator function, the table, starting point
    return stateless_iter, tbl, nil
end

function util.debug_table(tbl, lvl)
   lvl = lvl or 0
   local indent = ("   "):rep(lvl)
   local out = {"{\n"}
   for k,v in pairs(tbl) do
      table.insert(out, indent.."   ")
      table.insert(out, tostring(k).." = ")
      table.insert(out, (type(v) == "table" and util.debug_table(v, lvl + 1) or format_str(v)))
      table.insert(out, ",\n")
   end
   table.insert(out, indent.."}")
   return table.concat(out)
end

function util.collect_exec_data(job)
    local script = job.parameters.csbase_command_path.."/collect_execution_data"
    if not stat.stat(script) then
        return pairs({})
    end

    return coroutine.wrap(function ()
        local params = {
            job.cmd_id,
            job.sandboxes[1],
            job.parameters.csbase_command_path,
            job.parameters.csbase_command_output_path,
            job.parameters.csbase_command_root_path
        }
        local collect_cmd = "ksh "..script.." "..table.concat(params," ")
        --logger.debug("Executing data collection script: "..collect_cmd)
        local stdout = assert(io.popen(collect_cmd, 'r'))

        for line in stdout:lines() do
            for key, value in string.gmatch(line, "(.-)%s*=%s*([^,]+)") do
                coroutine.yield(key, value)
            end
        end
        stdout:close()
    end)
end

local function can_open(path, mode)
   local fd = io.open(path, mode)
   return fd and fd:close() or false
end

local function getPermissions(path)
    local tmpname = os.tmpname()
    local r = "test -r '%s' && echo -n r;"
    local w = "test -w '%s' && echo -n w;"
    local x = "test -x '%s' && echo -n x"

    local cmd = "("..r..w..x..") > %s"
    cmd = string.format(cmd, path, path, path, tmpname)

    os.execute(cmd)
    local file = io.open(tmpname, "r")
    if not file then
       local msg = "Failed to open command output file [%s]: %s."
       return nil, string.format(msg, tmpname, cmd)
    end
    local content = file:read("*a")
    file:close()
    os.remove(tmpname)

    local result = {}
    result.r = string.find(content, 'r') ~= nil
    result.w = string.find(content, 'w') ~= nil
    result.x = string.find(content, 'x') ~= nil
    return result
end

function util.list_local_path(path_name)
   local attrs = lfs.attributes(path_name)
   local list = {}
   if attrs and attrs.mode == "directory" then
      for file_name in lfs.dir(path_name) do
         if file_name ~= "." and file_name ~= ".." then
            local file_path
            if path_name:sub(-1) == "/" then
               file_path = path_name..file_name
            else
               file_path = path_name.."/"..file_name
            end
            list[#list + 1] = util.check_local_path(file_path)
         end
      end
      return list
   end	   
   return nil, "Could not list path"
end

function util.check_local_path(path_name)
   local symlink_attrs = lfs.symlinkattributes(path_name)
   if symlink_attrs then
      local size = symlink_attrs.size
      local is_dir = symlink_attrs.mode == "directory"
      -- in case of links, get data from the target file
      if symlink_attrs.mode == "link" then
         local attrs = lfs.attributes(path_name)
         if attrs then
            --print(util.debug_table(attrs))
            size = attrs.size
            is_dir = attrs.mode == "directory"
         end
      end
      --print(util.debug_table(symlink_attrs))
      local permissions = getPermissions(path_name) or {}
      return {
         path = path_name,
         size_bytes = size,
         is_dir = is_dir,
         is_symlink = symlink_attrs.mode == "link",
         is_readable = permissions.r or false,
         is_writable = permissions.w or false,
         is_executable = permissions.x or false,
         link_path = symlink_attrs.target or path_name,
      }
   end
   return nil, "Could not read path"
end

function util.convert_to_list(map, key)
   local i = 1
   local tab = {}
   while (map[key.."."..i]) do
      table.insert(tab, map[key.."."..i])
      i = i + 1
  end
  return tab
end

function util.removedir(path)
    for file in lfs.dir(path) do
        if file ~= "." and file ~= ".." then
            local f = path..'/'..file
            local attr = lfs.attributes(f)
            assert(attr)
            if attr.mode == "directory" then
                util.removedir(f)
            else
                local ok, err = os.remove(f)
                if not ok then
                    return ok, err
                end
            end
        end
    end
    return lfs.rmdir(path)
end

return util
